@@ -98,20 +98,25 @@ VERBS = {
     'gain':["gain", "gains"],
     'play_piece':["play", "plays"],
     'play_treasures':["play", "plays"],
-    'discard_piece':["discard", "discards"],
+    'discard':["discard", "discards"],
     'discard_pieces':["discard", "discards"],
-    'discard_zone':["discard", "discards"],
+    'discard_aside':["discard the cards you set aside earlier",
+                     "discards the cards they set aside earlier"],
     'reveal':["reveal", "reveals"],
+    'reveal_pieces':["reveal", "reveals"],
     'trash':["trash", "trashes"],
     'trash_pieces':["trash", "trashes"],
     'draw':["draw", "draws"],
     'do_nothing':["do nothing", "does nothing"],
     'topdeck':["topdeck", "topdecks"],
-    'put':["put", "puts"],
+    'put':["draw", "puts"],
     'set_aside':["set aside", "sets aside"],
-    'swap_top_two_cards':["swap the top two cards", "swaps the top two cards"],
+    'swap_top_cards_of_deck':["swap",
+                              "reverses the order of the top two cards of their deck"],
     'resolve_effects':["replay", "replays"],
     'resolve':["resolve", "resolves"],
+    'add_coin':["increase your coins by",
+                "increases their coins by"],
 }
 
 COMMAND_HINTS = {
@@ -405,21 +410,36 @@ def represent_effect_kwargs(kwargs):
         if (kwarg != "actor"):
             if (kwarg == "pieces"):
                 result.append(represent_pieces(kwargs[kwarg]))
-            elif (kwarg not in ["source", "destination", "state", "is_buy"]):
+            #elif (kwarg == "piece"):
+            #    result.append(represent_pieces(kwargs[kwarg]))
+            elif (kwarg not in ["source", "destination", "state", "is_buy", "free"]):
                 result.append(kwargs[kwarg])
     return commas(result)
 
 
 def represent_effect(effect, idx=0):
-    verb_string = '{}'.format(VERBS[effect.function.__name__][idx])
-    kwargs_string = represent_effect_kwargs(effect.kwargs)
+    funcname = effect.function.__name__
+    verb_string = '{}'.format(VERBS[funcname][idx])
+    # Case: Third person perspective, don't show drawn cards.
+    if (idx and (funcname == "put")):
+        kwargs_string = "a card into their hand"
+    # Case: First person perspective, show which cards are being
+    #       swapped by Sentry.
+    elif (not(idx) and (funcname == "swap_top_cards_of_deck")):
+        topcards = effect.kwargs['topcards']
+        kwargs_string = "{} to be the top card instead of {}".format(topcards[1],
+                                                                     topcards[0])
+    # Case: Default use.
+    else:
+        kwargs_string = represent_effect_kwargs(effect.kwargs)
     if kwargs_string:
         return "{} {}".format(verb_string, kwargs_string)
     return verb_string
 
 
 def represent_effects(effects, lowercase, idx):
-    effects = '; then '.join([represent_effect(effect, idx) for effect in effects])
+    effects_to_represent = list(filter(lambda effect: not(effect.__class__.__name__ == "Update"), effects))
+    effects = '; then '.join([represent_effect(effect, idx) for effect in effects_to_represent])
     effects = effects if lowercase else uppercase(effects)
     return add_period(effects)
 
