@@ -187,7 +187,7 @@ def draw(state, actor, n):
         actor.peek(actor.HAND[-true_n:])
 
 
-def swap_top_cards_of_deck(state, actor):
+def swap_top_cards_of_deck(state, actor, topcards):
     """
     Used by Sentry when there are at least two cards
     in actor's Deck, the top two cards have different
@@ -215,6 +215,10 @@ def reveal(state, actor, piece):
     pass
 
 
+def reveal_pieces(state, actor, pieces):
+    pass
+
+
 def do_nothing(state, actor):
     """
     Represent choices which have no impact on the game, such as
@@ -238,7 +242,7 @@ def topdeck(state, actor, piece, source):
     transfer_piece(piece, destination=actor.DECK, source=source)
 
 
-def discard_piece(state, actor, piece, source=None):
+def discard(state, actor, piece, source=None):
     source = actor.HAND if (source is None) else source
     to_discard, to_keep = partition(lambda x: (x is piece), source)
     source.clear()
@@ -259,7 +263,8 @@ def discard_pieces(state, actor, pieces, source=None):
     actor.DISCARD.extend(to_discard)
 
 
-def discard_zone(state, actor, source):
+def discard_aside(state, actor):
+    source = actor.ASIDE
     actor.DISCARD.extend(source)
     source.clear()
 
@@ -435,12 +440,22 @@ class Process:
     preventing the main Subprocess encoding the attack itself from ever
     unfolding.
     """
-    __slots__ = "actor", "main_effect", "setup_effect", "teardown_effect"
-    def __init__(self, actor, main_effect, setup_effect=None, teardown_effect=None):
+    __slots__ = "actor", "main_effect", "setup_effect", "teardown_effect", "count", "max_count"
+    def __init__(self, actor, main_effect, setup_effect=None, teardown_effect=None, max_count=0):
         self.actor = actor
         self.main_effect = main_effect
         self.setup_effect = setup_effect
         self.teardown_effect = teardown_effect
+        self.count = 0
+        self.max_count = max_count
+
+    @property
+    def count_antecedent(self):
+        """
+        Support antecedents which factor in the number of iterations
+        performed relative to the maximum allowed.
+        """
+        return self.count < self.max_count
 
     @property
     def antecedent(self):
@@ -449,9 +464,13 @@ class Process:
     def clear_internal_state(self):
         """
         Used by Processes which depend on internal state to compute
-        the antecedent.
+        the antecedent; by default, calibrates the main loop counter.
         """
-        pass
+        self.count = 0
+
+    def increment_count(self):
+        """ Hook for Subprocesses to directly control the number of iterations. """
+        self.count = self.count + 1
 
     def setup(self, state):
         if (self.setup_effect is not None):
@@ -646,30 +665,38 @@ class ImmunityProcedure(Subprocess):
 
 class Attack(Process):
     """ Processes which check for victim immunity. """
-    __slots__ = "immunity", "flag"
-    def __init__(self, actor, main_effect, teardown_effect=None):
+    __slots__ = "immunity"
+    def __init__(self, actor, main_effect, teardown_effect=None, max_count=1):
         super().__init__(actor=actor,
                          main_effect=main_effect,
                          setup_effect=Initiate(actor, ImmunityProcedure(self)),
-                         teardown_effect=teardown_effect)
+                         teardown_effect=teardown_effect,
+                         max_count=max_count)
         self.immunity = False
-        self.flag = False
 
     def mark_immunity(self):
         self.immunity = True
 
     def clear_internal_state(self):
+        super().clear_internal_state()
         self.immunity = False
-        self.flag = False
-
-    def kill(self):
-        """ Support one-shot Attacks. """
-        self.flag = True
 
     @property
     def antecedent(self):
         """ Subclasses check a conjunction of this antecedent with their own. """
-        return not(self.immunity or self.flag)
+        return not(self.immunity)
+
+
+class OneshotAttack(Attack):
+    """
+    Processes which check for victim immunity and only force victims
+    to perform a procedure once.
+    """
+    __slots__ = tuple()
+    @property
+    def antecedent(self):
+        """ Not immune and yet to have performed the victim procedure. """
+        return super().antecedent and super().count_antecedent
 
 
 
@@ -800,7 +827,7 @@ class MilitiaVictimProcedure(Subprocess):
         choices = []
         discardables = get_pieces(actor.HAND, unique=True)
         for discardable in discardables:
-            choices.append(Consequence(Effect(discard_piece,
+            choices.append(Consequence(Effect(discard,
                                               actor=actor,
                                               piece=discardable)))
         return choices
@@ -844,9 +871,17 @@ class ChapelChoices(Decision):
         choices = [NullOption(actor)]
         trashables = actor.HAND
         if trashables:
+            trash_choices = []
             combos = get_piece_combinations(trashables, kmin=1, kmax=4)
             for combo in combos:
-                choices.append(Consequence(Effect(trash_pieces, actor=actor, pieces=combo)))
+                # QoL: Sort the pieces within alphabetically.
+                combo = sorted(combo, key=lambda p: repr(p))
+                trash_choices.append(Consequence(Effect(trash_pieces, actor=actor, pieces=combo)))
+            # QoL: Sort the choices descending by size and within each
+            #      size, alphabetically.
+            trash_choices.sort(key=lambda c: (len(c.effects[0].kwargs['pieces']),
+                                              repr(c.effects[0].kwargs['pieces'][0])))
+            choices.extend(trash_choices)
         return choices
 
 
@@ -870,7 +905,7 @@ class VassalChoices(Decision):
         else:
             topcard = actor.DECK[-1]
             # The option of discarding the top card is always available.
-            choices.append(Consequence(Effect(discard_piece,
+            choices.append(Consequence(Effect(discard,
                                               actor=actor,
                                               piece=topcard,
                                               source=actor.DECK)))
@@ -889,13 +924,12 @@ class VassalChoices(Decision):
 
 class BureaucratProcedure(Subprocess):
     """
-    Each other player reveals a Victory card from their hand
-    and puts it onto their deck (or reveals a hand with no
-    Victory cards).
+    Reveal a Victory card from hand and topdeck it.
+    If you can't, reveal hand.
     """
     __slots__ = tuple()
     def generate_choices(self, state, actor):
-        signal = Update(self.process.kill)
+        signal = Update(self.process.increment_count)
         victory_cards = get_pieces(actor.HAND,
                                    unique=True,
                                    predicate=lambda x: x.is_victory)
@@ -905,8 +939,8 @@ class BureaucratProcedure(Subprocess):
                 choice.add(signal)
                 return [choice]
             else:
-                reveal_effects = [Effect(reveal, actor=actor, piece=card) for card in actor.HAND]
-                return [Consequence(*reveal_effects, signal)]
+                return [Consequence(Effect(reveal_pieces, actor=actor, pieces=actor.HAND),
+                                    signal)]
         else:
             choices = []
             for victory_card in victory_cards:
@@ -916,8 +950,12 @@ class BureaucratProcedure(Subprocess):
         return choices
 
 
-class BureaucratAttack(Attack):
-    """ Uses flag to ensure victims only perform the routine once. """
+class BureaucratAttack(OneshotAttack):
+    """
+    Each other player reveals a Victory card from their hand
+    and puts it onto their deck (or reveals a hand with no
+    Victory cards).
+    """
     __slots__ = tuple()
     def __init__(self, actor):
         super().__init__(actor=actor,
@@ -937,6 +975,7 @@ class BureaucratChoices(Decision):
             outcome = Consequence()
         for victim in actor.opponents:
             outcome.add(BureaucratAttack(victim))
+            outcome.custom_message = ["attack with Bureaucrat!", "attacks with Bureaucrat!"]
         return [outcome]
 
 
@@ -953,33 +992,47 @@ class MoneylenderChoices(Decision):
         return choices
 
 
-class PoacherProcedure(Decision):
+
+class PoacherProcedure(Subprocess):
     __slots__ = tuple()
     def generate_choices(self, state, actor):
+        signal = Update(self.process.increment_count)
+        choices = []
         discardables = get_pieces(actor.HAND, unique=True)
-        effects = [Effect(discard_piece, actor=actor, piece=card) for card in discardables]
-        return [Consequence(*effects)]
+        for discardable in discardables:
+            discard_effect = Effect(discard, actor=actor, piece=discardable)
+            choices.append(Consequence(discard_effect, signal))
+        return choices
+
+
+class PoacherProcess(Process):
+    __slots__ = tuple()
+    def __init__(self, actor, max_count):
+        super().__init__(actor=actor,
+                         main_effect=Initiate(actor, PoacherProcedure(self)),
+                         setup_effect=None,
+                         teardown_effect=None,
+                         max_count=max_count)
+
+    @property
+    def antecedent(self):
+        """
+        At least one card left to discard and at least one card available.
+        """
+        assert super().count_antecedent and len(self.actor.HAND)
 
 
 class PoacherChoices(Decision):
     """
     Discard a card per empty Supply pile.
     """
-    __slots__ = "procedure"
-    def __init__(self):
-        super().__init__()
-        self.procedure = PoacherProcedure()
-
+    __slots__ = tuple()
     def generate_choices(self, state, actor):
         n_to_discard = state.n_empty_piles
-        n_to_discard = min(n_to_discard, len(actor.HAND))
+        n_to_discard = min(len(actor.HAND), n_to_discard)
         if not(n_to_discard):
             return [NullOption(actor)]
-        else:
-            consequence = Consequence()
-            for i in range(n_to_discard):
-                consequence.add(Initiate(actor, self.procedure))
-            return [consequence]
+        return [Consequence(PoacherProcess(actor=actor, max_count=n_to_discard))]
 
 
 class ThroneRoomChoices(Decision):
@@ -1010,7 +1063,7 @@ class BanditProcedure(Subprocess):
         return piece.is_treasure and (repr(piece) != "Copper")
 
     def generate_choices(self, state, actor):
-        signal = Update(self.process.kill)
+        signal = Update(self.process.increment_count)
         prepare_deck(actor, n=2)
         if not(actor.DECK):
             option = NullOption(actor)
@@ -1020,14 +1073,15 @@ class BanditProcedure(Subprocess):
         decksize = len(actor.DECK)
         if (decksize == 1):
             topcard = actor.DECK[-1]
-            effect_function = trash if self.could_trash(topcard) else discard_piece
+            effect_function = trash if self.could_trash(topcard) else discard
             return [Consequence(Effect(reveal, actor=actor, piece=topcard),
                                 Effect(effect_function, actor=actor, piece=topcard, source=actor.DECK),
                                 signal)]
         else:
             choices = []
             topcards = [actor.DECK[-1], actor.DECK[-2]]
-            reveal_effects = [Effect(reveal, actor=actor, piece=card) for card in topcards]
+            #reveal_effects = [Effect(reveal, actor=actor, piece=card) for card in topcards]
+            reveal_effect = Effect(reveal_pieces, actor=actor, pieces=topcards)
             trashables, discardables = partition(self.could_trash, topcards)
             unique_trashables = get_pieces(trashables, unique=True)
             n_unique_trashables = len(unique_trashables)
@@ -1038,12 +1092,12 @@ class BanditProcedure(Subprocess):
                     trash_index, discard_index = index_permutation
                     trashable = unique_trashables[trash_index]
                     discardable = unique_trashables[discard_index]
-                    choices.append(Consequence(*reveal_effects,
+                    choices.append(Consequence(reveal_effect,
                                                Effect(trash,
                                                       actor=actor,
                                                       piece=trashable,
                                                       source=actor.DECK),
-                                               Effect(discard_piece,
+                                               Effect(discard,
                                                       actor=actor,
                                                       piece=discardable,
                                                       source=actor.DECK),
@@ -1052,14 +1106,14 @@ class BanditProcedure(Subprocess):
             #       or, both cards being trashable while sharing a piece type.
             elif (n_unique_trashables == 1):
                 trashable = unique_trashables[0]
-                consequence = Consequence(*reveal_effects,
+                consequence = Consequence(reveal_effect,
                                           Effect(trash,
                                                  actor=actor,
                                                  piece=trashable,
                                                  source=actor.DECK))
                 if discardables:
                     discardable = discardables[0]
-                    consequence.add(Effect(discard_piece,
+                    consequence.add(Effect(discard,
                                            actor=actor,
                                            piece=discardable,
                                            source=actor.DECK))
@@ -1067,7 +1121,7 @@ class BanditProcedure(Subprocess):
                 choices.append(consequence)
             # Case: Nothing to trash, discard both pieces.
             else:
-                choices.append(Consequence(*reveal_effects,
+                choices.append(Consequence(reveal_effect,
                                            Effect(discard_pieces,
                                                   actor=actor,
                                                   pieces=discardables,
@@ -1076,7 +1130,7 @@ class BanditProcedure(Subprocess):
         return choices
 
 
-class BanditAttack(Attack):
+class BanditAttack(OneshotAttack):
     __slots__ = tuple()
     def __init__(self, actor):
         super().__init__(actor=actor,
@@ -1093,32 +1147,44 @@ class BanditChoices(Decision):
         gold = state.filter_supply(lambda x: repr(x) == "Gold")
         if gold:
             outcome = Acquisition(Effect(gain, actor=actor, piece=gold[-1]))
+            custom_message = ["gain a Gold then attack with Bandit!",
+                              "gains a Gold then attacks with Bandit!"]
         else:
             outcome = Consequence(Effect(do_nothing, actor=actor))
+            custom_message = ["attack with Bandit!", "attacks with Bandit!"]
         for victim in actor.opponents:
             outcome.add(BanditAttack(victim))
+        outcome.custom_message = custom_message
         return [outcome]
 
 
-class CouncilRoomProcedure(Decision):
-    """
-    Each other player draws a card.
-    """
+class CouncilRoomProcedure(Subprocess):
+    """ Draw a card. """
     __slots__ = tuple()
     def generate_choices(self, state, actor):
-        return [Consequence(Effect(draw, actor=actor, n=1))]
+        return [Consequence(Effect(draw, actor=actor, n=1),
+                            Update(self.process.increment_count))]
+
+class CouncilRoomProcess(Process):
+    """ Each other player draws a card. """
+    __slots__ = tuple()
+    def __init__(self, actor):
+        super().__init__(actor=actor,
+                         main_effect=Initiate(actor, CouncilRoomProcedure(self)))
+
+    @property
+    def antecedent(self):
+        return self.count < self.max_count
 
 
 class CouncilRoomChoices(Decision):
-    __slots__ = "procedure"
-    def __init__(self):
-        super().__init__()
-        self.procedure = CouncilRoomProcedure()
-
+    __slots__ = tuple()
     def generate_choices(self, state, actor):
         interaction = Consequence()
         for victim in actor.opponents:
-            interaction.add(Initiate(victim, self.procedure))
+            interaction.add(CouncilRoomProcess(victim))
+        interaction.custom_message = ["prompt each other player to draw a card with Council Room.",
+                                      "prompts each other player to draw a card with Council Room."]
         return [interaction]
 
 
@@ -1147,9 +1213,10 @@ class LibraryTeardown(Subprocess):
     """ Discard any actions set aside during the Library Process. """
     __slots__ = tuple()
     def generate_choices(self, state, actor):
+        print("DEBUG | len(actor.HAND): {}".format(len(actor.HAND)))
         if not(actor.ASIDE):
             return [NullOption(actor)]
-        return [Consequence(Effect(discard_zone, actor=actor, source=actor.ASIDE))]
+        return [Consequence(Effect(discard_aside, actor=actor))]
 
 
 class LibraryProcess(Process):
@@ -1158,6 +1225,7 @@ class LibraryProcess(Process):
     LibraryProcedure so long as the preconditions hold.
     Afterwards, discard skipped cards using LibraryTeardown.
     """
+    __slots__ = tuple()
     def __init__(self, actor):
         super().__init__(actor=actor,
                          main_effect=Initiate(actor, LibraryProcedure(self)),
@@ -1175,10 +1243,16 @@ class LibraryProcess(Process):
 
 
 class LibraryChoices(Decision):
+    __slots__ = tuple()
     def generate_choices(self, state, actor):
         process = LibraryProcess(actor)
         if process.antecedent:
-            return [Consequence(process)]
+            consequence = Consequence(process)
+            consequence.custom_message = [
+                "begin the process described by Library.",
+                "begins the process described by Library.",
+             ]
+            return [consequence]
         return [NullOption(actor)]
 
 
@@ -1200,13 +1274,16 @@ class SentryChoices(Decision):
         decksize = len(actor.DECK)
         topcards = [actor.DECK[-1]] if (decksize == 1) else [actor.DECK[-1], actor.DECK[-2]]
 
+        non_null_choices = []
         # Do the same thing (trash or discard) to the topcard(s).
-        functions = [trash, discard_piece]
+        functions = [trash, discard]
         for function in functions:
             consequence = Consequence()
             for topcard in topcards:
                 consequence.add(Effect(function, actor=actor, piece=topcard, source=actor.DECK))
-            choices.append(consequence)
+            non_null_choices.append(consequence)
+        non_null_choices.sort(key=lambda c: (repr(c.effects[0].function.__name__)))
+        choices.extend(non_null_choices)
 
         unique_topcards = get_pieces(topcards, unique=True)
         n_unique_topcards = len(unique_topcards)
@@ -1216,23 +1293,30 @@ class SentryChoices(Decision):
         # Discarding one and leaving the other on top; or,
         # Leaving both on top but swapping their positions.
         if (n_unique_topcards == 2):
-            choices.append(Consequence(Effect(swap_top_cards_of_deck, actor=actor)))
+            non_null_choices = []
+            choices.append(Consequence(Effect(swap_top_cards_of_deck,
+                                              actor=actor,
+                                              topcards=unique_topcards)))
             index_permutations = [(0, 1), (1, 0)]
             for index_permutation in index_permutations:
                 trash_index, discard_index = index_permutation
-                choices.append(Consequence(Effect(trash, actor=actor, piece=topcards[trash_index], source=actor.DECK),
-                                           Effect(discard_piece, actor=actor, piece=topcards[discard_index], source=actor.DECK)))
-                choices.append(Consequence(Effect(trash, actor=actor, piece=topcards[trash_index], source=actor.DECK)))
-                choices.append(Consequence(Effect(discard_piece, actor=actor, piece=topcards[discard_index], source=actor.DECK)))
+                non_null_choices.append(Consequence(Effect(trash, actor=actor, piece=topcards[trash_index], source=actor.DECK),
+                                           Effect(discard, actor=actor, piece=topcards[discard_index], source=actor.DECK)))
+                non_null_choices.append(Consequence(Effect(trash, actor=actor, piece=topcards[trash_index], source=actor.DECK)))
+                non_null_choices.append(Consequence(Effect(discard, actor=actor, piece=topcards[discard_index], source=actor.DECK)))
+            non_null_choices.sort(key=lambda c: (repr(c.effects[0].function.__name__)))
+            choices.extend(non_null_choices)
 
+        # QoL: Sort the order choices are displayed by number of cards involved.
+        choices.sort(key=lambda c: len(c.effects))
         return choices
 
 
 class WitchProcedure(Subprocess):
-    """ Each other player gains a Curse. """
+    """ Gain a Curse if able. """
     __slots__ = tuple()
     def generate_choices(self, state, actor):
-        signal = Update(self.process.kill)
+        signal = Update(self.process.increment_count)
         gainables = state.filter_supply(lambda x: repr(x) == "Curse")
         if gainables:
             option = Acquisition(Effect(gain, actor=actor, piece=gainables[0]),
@@ -1243,18 +1327,21 @@ class WitchProcedure(Subprocess):
         return [option]
 
 
-class WitchAttack(Attack):
+class WitchAttack(OneshotAttack):
+    """ Each other player gains a Curse. """
     __slots__ = tuple()
     def __init__(self, actor):
         super().__init__(actor=actor,
                          main_effect=Initiate(actor, WitchProcedure(self)))
 
 
+
 class WitchChoices(Decision):
     def generate_choices(self, state, actor):
         outcome = Consequence()
         for victim in actor.opponents:
             outcome.add(WitchAttack(victim))
+        outcome.custom_message = ["attack with Witch!", "attacks with Witch!"]
         return [outcome]
 
 
@@ -1280,11 +1367,11 @@ class ArtisanChoices(Decision):
                 for placeable in placeables:
                     choices.append(Acquisition(Effect(gain, actor=actor, piece=gainable, destination=actor.HAND),
                                                Effect(topdeck, actor=actor, piece=placeable, source=actor.HAND)))
-                    # Case: Gaining a Piece with a novel type relative to Pieces in Hand
-                    #       implies the option to place a Piece of that type has not yet
-                    #       been generated (otherwise it will have been, so avoid duplication).
-                    if (gainable_toi not in placeable_tois):
-                        choices.append(Acquisition(Effect(gain, actor=actor, piece=gainable, destination=actor.HAND),
-                                                   Effect(topdeck, actor=actor, piece=gainable, source=actor.HAND)))
+                # Case: Gaining a Piece with a novel type relative to Pieces in Hand
+                #       implies the option to place a Piece of that type has not yet
+                #       been generated (otherwise it will have been, so avoid duplication).
+                if (gainable_toi not in placeable_tois):
+                    choices.append(Acquisition(Effect(gain, actor=actor, piece=gainable, destination=actor.HAND),
+                                               Effect(topdeck, actor=actor, piece=gainable, source=actor.HAND)))
 
         return choices
